#!/usr/bin/perl
#
# Matt's Web Page Editor v1.90
# Copyright (c) 1995 Matthew Miller

#----------- Part One: Set some essential variables
  
  $configfile = "mmwebedit.config";

#----------- Part Two: Read in config file information
#   
#                      Temp Variables in this section:
#                        @configlines,$configfile hold the config data
#                        $name and $value are temporary bits from the each line
#                      We end up with
#                        @config{'keyname'} to access config data keys, and
#                        $maindir, which is actually $config{'maindir'} 
#                        $mmwebedit, which is actually $config{'mmwebedit'}

  #open the config file, dump it into the configlines array, close the file
  open (CONFIGFILE, "$configfile");
  @configlines = <CONFIGFILE>;
  close (CONFIGFILE);

  #read the different lines of the config file, split out the different
  #keys, and ignore comment line that begin with #. The chop statement
  #is to discard the extra linefeed.
  foreach $configline (@configlines)
  {
      if (substr($configline,0,1) ne "#")
      { 
        chop ($configline);
        ($name, $value) = split (/=/, $configline);
        $config{$name}=$value;
      }
  }

 #one important setting is the main directory, so we'll read it into it's
 #own variable for convenience
 $maindir = $config{'maindir'};
 $mmwebedit = $config{'mmwebedit'};
  
#----------- Part Three: If there's no input at all, send them to the login screen 
#
#                      No Variables in this section.
 
  if (!$ENV{'CONTENT_LENGTH'})
  {
      print "Location: $config{'loginform'}\n\n";
      exit;
  }
  
              
#----------- Part Four: Process input
#
#                        Temp Variables in this section:
#                          $buffer  is the incoming data
#                          @forminputs, $forminput are places to stick each line of the data
#                          $name, $value are bits of each line
#                        And we end up with
#                          @results{'inputname'} with the results of the form.


  #read the incoming data
  read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});

  #seperate the different form inputs.
  @forminputs = split(/&/, $buffer);

  #for each line, seperate out the value from the name of the input control
  foreach $forminput (@forminputs)
  {
      #seperate the name of the the input from the value
      ($name, $value) = split (/=/, $forminput);

      #Un-webify plus signs and %-encoding
      $value =~ tr/+/ /;
      $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
  
      #and then stick them in the results array so we can look at 'em
      #by refering to the name of the input form
      $results{$name}=$value;


      #passwords can be set to always be lowercased--effectively makes them
      #case insensitive, UNLESS the original password was entered into the
      #password file with upper case letters, in which case this will cause
      #the password to NEVER work. Use the config file to change this.
      if ($config{'lowercasepasswords'} eq "yes")
      {
         $results{'password'} =~ tr/A-Z/a-z/;
      }  
     
  }
  
#----------- Part Five: Read in the usernames and passwords.
#
#                       Temp Variables in this section:
#                           @people, $person are data from the password file
#                           $username, $password, $firstname, $lastname, $other
#                             are all bits of password file data
#                       And we end up with these useful arrays:
#                           @firstnames{'username'}
#                           @lastnames{'username'}
#                           @passwords{'username'}
#
# Currently the password file entries are divided by spaces. This should
# be changed to allow for spaces in names. Maybe = or +.

                      
  #read the password file into the people array
  open (PEOPLE, $config{'passwordfile'});
  @people = <PEOPLE>;
  close (PEOPLE);

  #split out the different bits of the password file into
  #various arrays indexed by the username
  foreach $person (@people)
  {
     #discard linefeed and split the data
     chop $person;
     ($username, $password, $firstname, $lastname, $other) = split(/ /,$person);
     
     #lowercase the username
     $username =~ tr/A-Z/a-z/;

     #stick the data in useful arrays.
     $firstnames{$username}=$firstname;
     $lastnames{$username}=$lastname;
     $passwords{$username}=$password;
   }


#----------- Part Six: Process validation keys and usernames & passwords
#
#                       Temp Variables in this section:
#                            $givenusername, $givenvalidation are parts of the 
#                                                  given validation key 
#                            $unscrambledpassword is the unscrambled password
#                                                  from that part of the key
#                            @errormessage, $errorline
#
#                       We end up with 
#                            $correctusername
#                            $correctpassword
#
#                       If there is a validation key, check that it's
#                       valid. If not, give the login form. (An invalid
#                       verification key probably means serious problems
#                       with mmwebedit.)
#
#                       If there is no validation key, check for the 
#                       username and password. If they're there but not
#                       correct, give the proper error mesage.
#                        
#                       If neither validation nor username/password exist,
#                       then we'll assume that the username and password
#                       were there but both blank. I'm too lazy to deal with
#                       the possiblity of other weird cases which would only
#                       happen if something were really screwed up, in which
#                       case giving the "no username given" error message is
#                       still okay.
#
#      Anyway, the end result of all that is that the script will die 
#      somewhere in here if a correct password and username are not given.

  if ($results{'validation'})
  {
       #split the validation key on the space
       ($givenusername, $givenvalidation) = split (/ /, $results{'validation'});
 
       #check if the validation key matches the password file. The validation
       #key is the password with the letters scrambled a bit, so that it
       #isn't obvious if someone views the HTML source of the editor screen.
       #I can't really think of a better way to do it than that.
 
       $unscrambledpassword = $givenvalidation;
       $unscrambledpassword =~ tr/[N-Zn-z5-9][A-Ma-m0-4]/[A-Ma-m0-4][N-Zn-z5-9]/;
     
       if (!$passwords{$givenusername} || crypt($unscrambledpassword,"MM") ne $passwords{$givenusername})
       {
           #bad validation key means go to the login screen
           print "Location: $config{'loginform'}\n\n";
           exit;
       }    
       #else we're okay.

       $correctusername = $givenusername;
       $correctpassword = $unscrambledpassword;
       
       #usernames aren't case sensitive
       $correctusername =~ tr/A-Z/a-z/;  
   }
   else #no verification key found. assume username/password
   {     
       if (!$results{'username'})
       {   #No username given
           print "Location: $config{'error_nousername'}\n\n";
           exit;
       }
       elsif (!$passwords{$results{'username'}})
       {   #Bad username given
           open (ERRORMESSAGE, $config{'error_badusername'});
           @errormessage = <ERRORMESSAGE>;
           close (ERRORMESSAGE);
 
           #CGI Header
           print "Content-type: text/html\n\n";

           #Translate username variable in error message, then print
           foreach $errorline (@errormessage)
           {
             $errorline =~s/\$username/$results{'username'}/gi;
             print "$errorline\n";
           }

           exit;
         
       }
       elsif(crypt($results{'password'},"MM") ne $passwords{$results{'username'}})
       {  #Bad password given

           open (ERRORMESSAGE, $config{'error_badpassword'});
           @errormessage = <ERRORMESSAGE>;
           close (ERRORMESSAGE);
 
           #CGI Header
           print "Content-type: text/html\n\n";

           #Translate username variable in error message, then print
           foreach $errorline (@errormessage)
           {
             $errorline =~s/\$username/$results{'username'}/gi;
             print "$errorline\n";
           }

           exit;
       }
   
       $correctusername = $results{'username'};
       $correctpassword = $results{'password'};
       $correctusername =~ tr/A-Z/a-z/; #usernames aren't case sensitive
 
   }

#----------- Part Seven: generate validation key
#
#                      Temp variables:
#                        $scrambledpassword                    
#                      We end up with 
#                        $validationkey

    $scrambledpassword = $correctpassword;
    $scrambledpassword =~ tr/[A-Ma-m0-4][N-Zn-z5-9]/[N-Zn-z5-9][A-Ma-m0-4]/;
    $validationkey = $correctusername . " " . $scrambledpassword;


#----------- Part Eight: Generate actual pagename
#
#                     Temp Variables in the section:
#                        $temppagename
#                        $userdirname 
#                     We end up with:
#                        $correctpagename name of the page without 
#                                 the extension or anything
#                                 (variable name chosen because it
#                                  matches $correctusername and $correctpassword
#                        $pagename  name o' the page with path info
#                        $pageurl   complete URL of the page

   #if their directory doesn't exist, die a nasty twitching death.
   $userdirname = "$maindir" . "$correctusername";
   if (!-d $userdirname)
   {   
       print "Location: $config{'error_nodirectory'}\n\n";
       exit;
   }


   #if there is no pagename, we use index.html
   if (!$results{'pagename'}) 
   {
       $correctpagename="index";
       $pagename=$maindir . $correctusername . "/index.html";
       $pageurl=$config{'mainurl'} . $correctusername . "/";
   }
   else
   {
      #there's a given pagename. make sure it's all alphanumeric.
      $correctpagename=$results{'pagename'};
      $correctpagename =~ s/[^a-zA-Z0-9\-]/_/g;
      if ($correctpagename eq "" || $correctpagename eq "index")
      {
         $correctpagename="";
         $pagename=$maindir . $correctusername . "/" . "index.html";
         $pageurl=$config{'mainurl'} . $correctusername . "/";
      }
      else
      { 
        $pagename=$maindir . $correctusername . "/" . $correctpagename . ".html";
        $pageurl=$config{'mainurl'} . $correctusername . "/" . $correctpagename . ".html";
      }

   }


#----------- Part Nine : If there's a switchpage key, switch the page
#
#

   if ($results{'switchpage'})
   { 
      $correctpagename=$results{'switchpage'};
      $correctpagename =~ s/[^a-zA-Z0-9\-]/_/g;

      if ($correctpagename eq "" || $correctpagename eq "index")
      {
         $correctpagename="";
         $pagename=$maindir . $correctusername . "/" . "index.html";
         $pageurl=$config{'mainurl'} . $correctusername . "/";
      }
      else
      { 
        $pagename=$maindir . $correctusername . "/" . $correctpagename . ".html";
        $pageurl=$config{'mainurl'} . $correctusername . "/" . $correctpagename . ".html";
      }

   }


#----------- Part Ten: add new page if so requested


   if ($results{'createpage'})
   { 
      $correctpagename=$results{'createpage'};
      $correctpagename =~ s/[^a-zA-Z0-9\-]/_/g;

      if ($correctpagename eq "" || $correctpagename eq "index")
      {
         $correctpagename="";
         $pagename=$maindir . $correctusername . "/" . "index.html";
         $pageurl=$config{'mainurl'} . $correctusername . "/";
      }
      else
      { 
        $pagename=$maindir . $correctusername . "/" . $correctpagename . ".html";
        $pageurl=$config{'mainurl'} . $correctusername . "/" . $correctpagename . ".html";

        if (!-e $pagename)
        {
          if ($results{'usetemplate'} eq "yes")
          { 
             open (TEMPLATEFILE, $config{'template'});
             @templatefile = <TEMPLATEFILE>;
             close (TEMPLATEFILE);
          }
          else
          { 
             @templatefile = "";
          }
 
          open (NEWFILE, ">$pagename");
          print NEWFILE @templatefile;
          close (NEWFILE);

        }
      }
   }
#----------- Part Eleven: rename a file if we're supposed to
#
#
#

   if ($results{'renamepage'})
   { 
      $renamepagename=$results{'renamepage'};
      $renamepagename =~ s/[^a-zA-Z0-9\-]/_/g;

      $originalpagename =$results{'originalpage'};
      $originalpagename =~ s/[^a-zA-Z0-9\-]/_/g;

      $renamefilename=$maindir . $correctusername . "/" . $renamepagename . ".html";
      $originalfilename=$maindir . $correctusername . "/" . $originalpagename . ".html";

        if (!-e $renamefilename && -e $originalfilename)
        {           
           rename ($originalfilename, $renamefilename);
        }      

      $correctpagename=$renamepagename;
      $pagename=$maindir . $correctusername . "/" . $correctpagename . ".html";
      $pageurl=$config{'mainurl'} . $correctusername . "/" . $correctpagename . ".html";

   }


#----------- Part Twelve: Delete a file if we're supposed to
#
#
#

   if ($results{'deletepage'})
   { 
      $deletepagename=$results{'deletepage'};
      #we'll have no non-alphanumerics--no changing directories or anything
      #sneaky like that.
      $deletepagename =~ s/[^a-zA-Z0-9\-]/_/g;

        $deletefilename=$maindir . $correctusername . "/" . $deletepagename . ".html";

        if (-e $deletefilename)
        {           
           unlink $deletefilename;
        }      
   }

#----------- Part Thirteen: read in directory information
#
#                  Temp Variables: 
#                         $userdirname            
#                         @dirlines, $dirline
#                  Final:
#                         @usershtmlfiles

   $userdirname = "$maindir" . "$correctusername";

   opendir (USERDIR, "$userdirname");
   @dirlines = readdir (USERDIR);
   closedir (USERDIR);

   foreach $dirline (@dirlines)
   { 
    if (substr($dirline,length($dirline)-5) eq ".html")
    {
      if ($dirline ne "index.html")
      {
      push (@usershtmlfiles, substr($dirline,0,length($dirline)-5));
      }
    }
   }


#----------- Part Fourteen: FileManager Form

   if ($results{'fileman'})
   {

      #read in the HTML instructions for the file manager
      open (MMWFHTML1, $config{'mmwf-html1'});
      @mmwfhtml1 = <MMWFHTML1>;
      close (MMWFHTML1);

      open (MMWFHTML2, $config{'mmwf-html2'});
      @mmwfhtml2 = <MMWFHTML2>;
      close (MMWFHTML2);

      open (MMWFHTML3, $config{'mmwf-html3'});
      @mmwfhtml3 = <MMWFHTML3>;
      close (MMWFHTML3);

      open (MMWFHTML4, $config{'mmwf-html4'});
      @mmwfhtml4 = <MMWFHTML4>;
      close (MMWFHTML4);

      open (MMWFHTML5, $config{'mmwf-html5'});
      @mmwfhtml5 = <MMWFHTML5>;
      close (MMWFHTML5);

      open (MMWFHTML6, $config{'mmwf-html6'});
      @mmwfhtml6 = <MMWFHTML6>;
      close (MMWFHTML6);

     #CGI Header
     print "Content-type: text/html\n\n";
     
     #Directory of files:
       print @mmwfhtml1;
       
       print "<PRE>\n";
       print "<A HREF=\"$config{'mainurl'}$correctusername/\">(Default)</A>\n";
       foreach $usershtmlfile (@usershtmlfiles)
       {
          print "<A HREF=\"$config{'mainurl'}$correctusername/$usershtmlfile.html\">$usershtmlfile.html</A>\n";
       }
       print "</PRE>\n";

     #edit file
       print @mmwfhtml2;

       print "<FORM METHOD=post ACTION=\"$mmwebedit\">\n";
       print "<TABLE>\n";
       print "<INPUT NAME=validation TYPE=hidden VALUE=\"$validationkey\">\n";
       
       print "<TR>\n<TD>\n";
       print "<INPUT TYPE=submit VALUE=\"Edit File:\">\n";
       print "</TD><TD>\n";

       print "<SELECT NAME=pagename>\n";
       print "  <OPTION VALUE=\"index\">(Default)\n";
       
       foreach $usershtmlfile (@usershtmlfiles)
       {
          if ($usershtmlfile eq $correctpagename)
          {
             print "  <OPTION SELECTED>$usershtmlfile\n";
          }
          else
          {      
             print "  <OPTION>$usershtmlfile\n";
          }
       }
       print "</SELECT>";
       print "</TD>\n</TR>\n</TABLE>\n";
       print "</FORM>\n";

     #create file
       print @mmwfhtml3;
  

       print "<FORM METHOD=post ACTION=\"$mmwebedit\">\n";
       print "<TABLE>\n";

       print "<INPUT NAME=validation TYPE=hidden VALUE=\"$validationkey\">\n";
       print "<INPUT NAME=fileman TYPE=hidden VALUE=yes>\n";
       
       print "<TR>\n<TD>\n";
       print "<INPUT TYPE=submit VALUE=\"Create File:\">\n";

       print "</TD><TD>\n";

       print "<INPUT NAME=createpage SIZE=30>\n";
   
       print "</TD><TD>\n";
       print "<B>.html</B>";
       
       print "</TD>\n</TR>\n";
       print "<TD></TD><TD>\n";
       print "<INPUT TYPE=radio NAME=usetemplate VALUE=no CHECKED><B>Blank</B>";
       print "<INPUT TYPE=radio NAME=usetemplate VALUE=yes><B>Basic Template</B>\n";
       print "</TABLE>\n";
       print "</FORM>\n";

     #renamefile
       print @mmwfhtml4; 

       print "<FORM METHOD=post ACTION=\"$mmwebedit\">\n";
       print "<TABLE>\n";

       print "<INPUT NAME=validation TYPE=hidden VALUE=\"$validationkey\">\n";
       print "<INPUT NAME=pagename TYPE=hidden VALUE=\"$correctpagename\">\n";
       print "<INPUT NAME=fileman TYPE=hidden VALUE=yes>\n";

       print "<TR>\n<TD>\n";
       print "<INPUT TYPE=submit VALUE=\"Rename File:\">\n";

       print "</TD><TD>\n";

       print "<SELECT NAME=originalpage>\n";
       
       foreach $usershtmlfile (@usershtmlfiles)
       {
          if ($usershtmlfile eq $correctpagename)
          {
             print "  <OPTION SELECTED>$usershtmlfile\n";
          }
          else
          {      
             print "  <OPTION>$usershtmlfile\n";
          }
       }
       print "</SELECT>\n";
       print "</TD><TD>\n";
       print "<B>to</B> ";
       print "</TD><TD>\n";

       print "<INPUT NAME=renamepage SIZE=30>\n";
       print "</TD><TD>\n";
       print "<B>.html</B>";
       print "</TD><TD>\n</TR>\n</TABLE>\n";
       print "</FORM>\n";


     #delete file
       print @mmwfhtml5; 

       print "<FORM METHOD=post ACTION=\"$mmwebedit\">\n";
       print "<TABLE>\n";
       print "<INPUT NAME=validation TYPE=hidden VALUE=\"$validationkey\">\n";
       print "<INPUT NAME=fileman TYPE=hidden VALUE=yes>\n";

       print "<TR>\n\<TD>\n";
       print "<INPUT TYPE=submit VALUE=\"Delete File:\">\n";
       print "</TD><TD>\n";
       print "<SELECT NAME=deletepage>\n";
       
       foreach $usershtmlfile (@usershtmlfiles)
       {
          if ($usershtmlfile eq $correctpagename)
          {
             print "  <OPTION SELECTED>$usershtmlfile\n";
          }
          else
          {      
             print "  <OPTION>$usershtmlfile\n";
          }
       }
       print "</SELECT>\n";

       print "</TD><TD>\n</TR>\n</TABLE>\n";

       print "</FORM>\n";

     #footer
     print @mmwfhtml6;

 
     exit;
   }
        
#----------- Part Fifteen: Save to disk if needed
#
#                       No temp variables used:
#
#             If we got here from the editor form, there's probably data that
#             needs to be saved. If we got here from somewhere else, DON'T
#             save. We tell if a save is needed by the existence of a 
#             savepage key. Any value will work--just being there makes
#             the save happen.
# 
#             When we save to disk, any "[/textarea]"s are changed back to
#             "</textarea>"s. This is part of a work-around to avoid a 
#             major bug if we print a "</textarea>" in the middle of our
#             textarea form. This method could be used for implementing
#             some more advanced features (ready-made pages sections, etc.)

      
   #If there is a savepage key, save the page. Otherwise don't.
   if ($results{'savepage'})
   { 

      #[/textarea] conversion.   (doesn't that make a pretty search pattern...)
      $results{'pagetext'} =~ s/\[\/textarea\]/<\/textarea>/gi;

      open (HTMLFILE, ">$pagename");
      print HTMLFILE $results{'pagetext'};
      close (HTMLFILE);
   }


#----------- Part Sixteen: Read in the edited page to display it
#
#                       Temp variable:
#                         $htmlline
#                       We'll end up with
#                         @finalhtml   <-- the users' page
#                         @displayhtml <-- with </textarea> converted to 
#                                               [/textarea]   

    open(HTMLFILE, "$pagename");
    @finalhtml = <HTMLFILE>;
    close(HTMLFILE);

    @displayhtml=@finalhtml;

    foreach $htmlline (@displayhtml)
    { #convert </textarea> to [/textarea] for safe display
     $htmlline =~ s/<\/textarea>/\[\/textarea\]/gi;
    }
   

#----------- Part Seventeen: Read in the form header and footer
# 
#                       Temp variables in this section:
#                         $formheadline, $formfootline
#                       And we get:
#                         @formheader  
#                         @formfooter
#

 
   #read in the HTML that goes before and after the editing form. 
   #before:
   open (FORMHEAD, $config{'formheader'});
   @formheader = <FORMHEAD>;
   close (FORMHEAD);
 
   #after:
   open (FORMFOOT, $config{'formfooter'});
   @formfooter = <FORMFOOT>;
   close (FORMFOOT);

   #Translate name variables in the header and footer.
   foreach $formheadline (@formheader)
   {
     $formheadline =~s/\$firstname/$firstnames{$correctusername}/gi;
     $formheadline =~s/\$lastname/$lastnames{$correctusername}/gi;
     $formheadline =~s/\$username/$correctusername/gi;
     $formheadline =~s/\$pageurl/$pageurl/gi;
   }
   foreach $formfootline (@formfooter)
   {
     $formfootline =~s/\$firstname/$firstnames{$correctusername}/gi;
     $formfootline =~s/\$lastname/$lastnames{$correctusername}/gi;
     $formfootline =~s/\$username/$correctusername/gi;
     $formfootline =~s/\$pageurl/$pageurl/gi;
   }



   
#----------- Part Eighteen: display the actual HTML output 
#
#

   #CGI Header
   print "Content-type: text/html\n\n";
   print "<BASE HREF=\"$config{'mainurl'}$correctusername/\">";


   #preview page if it's supposed to be here
   if ($config{'previewbeforeedit'} eq "yes")
   {
     print @finalhtml;
     print "\n<P>\n";
   }

   #Form header from file
   print @formheader;

   #Page editor form
   print "<FORM METHOD=post ACTION=\"$mmwebedit\">\n";
   print "<TEXTAREA NAME=pagetext ROWS=$config{'formrows'} COLS=$config{'formcols'}>\n";

   #Page data so it can be edited
   print @displayhtml;

   print "</TEXTAREA><p>\n";
   print "<INPUT NAME=validation TYPE=hidden VALUE=\"$validationkey\">\n";
   print "<INPUT NAME=pagename TYPE=hidden VALUE=\"$correctpagename\">\n";
   print "<INPUT NAME=savepage TYPE=hidden VALUE=yes>\n";

   #i don't know if this table trick is legal HTML, but it works in Netscape.
   #(because the table and the forms overlap)
   if ($config{'usewackytables'} eq "yes")
   {
       print "<TABLE BORDER>\n";
       print "<TR>\n<TD>\n";   
   }

   print "<INPUT TYPE=submit VALUE=\"Enter Changes\">\n";
   
   #more possibly illegal form/table stuff
   if ($config{'usewackytables'} eq "yes")
   {
      print "</TD>\n";
   }

   print "</FORM>\n\n";

   #illegal?
   if ($config{'usewackytables'} eq "yes")
   {
      print "<TD>\n";
   }

   #Select Page form.
   print "<FORM METHOD=post ACTION=\"$mmwebedit\">\n";
   print "<INPUT NAME=validation TYPE=hidden VALUE=\"$validationkey\">\n";
   print "<INPUT NAME=pagename TYPE=hidden VALUE=\"$correctpagename\">\n";

   print "<INPUT TYPE=submit VALUE=\"Switch to:\">\n";
   print "<SELECT NAME=\"switchpage\">\n";
   print "  <OPTION VALUE=\"index\">(Default)\n";

   foreach $usershtmlfile (@usershtmlfiles)
   {	
      if ($usershtmlfile)
      {
        if ($usershtmlfile eq $correctpagename)
        {
           print "  <OPTION SELECTED>$usershtmlfile\n";
        }
        else
        {      
           print "  <OPTION>$usershtmlfile\n";
        }
      }
   }    
   print "</SELECT>\n";
   print "<B>without saving</B>\n";
   

   #illegal?
   if ($config{'usewackytables'} eq "yes")
   {
      print "<TD>\n";
   }

   print "</FORM>\n";

   #Filemanager form.
   print "<FORM METHOD=post ACTION=\"$mmwebedit\">\n";
   print "<INPUT NAME=validation TYPE=hidden VALUE=\"$validationkey\">\n";
   print "<INPUT NAME=pagename TYPE=hidden VALUE=\"$correctpagename\">\n";
   print "<INPUT NAME=fileman TYPE=hidden VALUE=yes>\n";

   print "<INPUT TYPE=submit VALUE=\"Add/Del Files\">\n";
   #and the last possibly illegal table stuff
   if ($config{'usewackytables'} eq "yes")
   {   
      print "</TD>\n</TR>\n</TABLE>\n";
   }

   print "</FORM>\n";

   print "<P>\n";
   
   #footer from file
   print @formfooter;
   
   #preview page if it's supposed to be here
   if ($config{'previewbeforeedit'} ne "yes")
   {
     print @finalhtml;
   }


#----------- The end.

